/*
 * out_yasapi.c
 * Copyright (C) 2015 Peter Belkner <pbelkner@snafu.de>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA
 */
#include <yasapi.h>

#define GETTIME(time) \
  (audioclock?time:GetTickCount())

int getwrittentime();
int getoutputtime();

static int state=0;
static int srate, numchan, bps;
static volatile int writtentime, w_offset;
static int start_t;
static int last_pause=0;
static int video;
static int audioclock;

static wchar_t *path;
static Plugin plugin;
static PluginOptions *pOptions=&plugin.options;
static Queue *pQueue=&plugin.queue;
static Store *pStore=&plugin.store;

#if 0 // {
BOOL WINAPI _DllMainCRTStartup(HANDLE hInst, ULONG ul_reason_for_call,
    LPVOID lpReserved)
{
  return TRUE;
}
#endif // }

static int GetAudioTime(void)
{
  Result *pResult;
  Request *pRequest;
  int time;

  if (1<state) {
    pResult=StoreGet(pStore);
    pRequest=QueueLockWrite(pQueue,pResult,REQUEST_GETTIME);
    QueueUnlockWrite(pQueue,pResult);
      time=pResult->state;
    StorePut(pStore,pResult);
  }
  else
    time=0;

  return time;
}

void config(HWND hwnd)
{
  HINSTANCE hInstance=winampGetOutModule()->hDllInstance;

  if (0<state&&plugin.pPlayer)
    ConfigDialog(hwnd,hInstance,plugin.pPlayer);
}

void about(HWND hwnd)
{
  AboutDialog(hwnd);
}

void init()
{
  int debug;

  if (NULL==(path=wapath(YASAPI_PROPERTY_FILE)))
    goto path;

#if defined (YASAPI_DEBUG) // {
  debug=GetPrivateProfileIntW(
    YASAPI_GROUP_GENERAL,     // _In_  LPCTSTR lpAppName,
    YASAPI_PROPERTY_DEBUG,    // _In_  LPCTSTR lpKeyName,
    PluginOptionsDefault()->debug,
                              // _In_  INT nDefault,
    path                      // _In_  LPCTSTR lpFileName
  );

  if (debug)
    ConsoleOpen();
#else // } {
  debug=0;
#endif // }

  DPRINTF(debug,0,"%s\n",__func__);
  DPRINTF(debug,0,"  module id: %d (%d)\n",YASAPI_WA_ID,
      winampGetOutModule()->id);

  if (PluginCreate(&plugin,path,debug)<0)
    goto plugin;

  ++state;

  return;
// cleanup:
  //PluginDestroy(&plugin);
plugin:
  GlobalFree(path);
path:
  return;
}

void quit()
{
  int debug;

  if (0<state) {
    --state;
    DPRINTF(pOptions->debug,0,"%s\n",__func__);

    PluginDestroy(&plugin);
    GlobalFree(path);
  }
}

int open(int samplerate, int numchannels, int bitspersamp, int bufferlenms,
    int prebufferms)
{
  Result *pResult;
  Request *pRequest;
  int maxlatency;

  DPRINTF(pOptions->debug,0,"%s\n",__func__);

  if (state<1)
    return -1;

  audioclock=pOptions->bAudioClock;
  DPRINTF(pOptions->debug,0,"  %s clock\n",audioclock?"wasapi":"system");

  w_offset = writtentime = 0;
  start_t = GETTIME(0);
  numchan = numchannels;
  srate = samplerate;
  bps = bitspersamp;

  last_pause=0;
  video=-666==bufferlenms;

  pResult=StoreGet(pStore);
  pRequest=QueueLockWrite(pQueue,pResult,REQUEST_OPEN);
    pRequest->open.samplerate=samplerate;
    pRequest->open.numchannels=numchannels;
    pRequest->open.bitspersamp=bitspersamp;
  QueueUnlockWrite(pQueue,pResult);
    maxlatency=pResult->state;
  StorePut(pStore,pResult);

  if (0<=maxlatency)
    ++state;

  return maxlatency;
}

void close()
{
#if 0 // {
  HWND hWndMain=winampGetOutModule()->hMainWindow;
  int fullstop=!SendMessageW(hWndMain,WM_WA_IPC,0,IPC_ISFULLSTOP);
#endif // }
  Result *pResult;
  Request *pRequest;

  DPRINTF(pOptions->debug,0,"%s\n",__func__);

  if (1<state) {
    --state;
    pResult=StoreGet(pStore);
    pRequest=QueueLockWrite(pQueue,pResult,REQUEST_CLOSE);
#if 0 // {
      pRequest->close.fullstop=fullstop;
#else // } {
      pRequest->close.fullstop=0;
#endif // }
    QueueUnlockWrite(pQueue,pResult);
    StorePut(pStore,pResult);
  }
}

int write(char *buf, int len)
{
  int failure;
  Result *pResult;
  Request *pRequest;

  DPRINTF(pOptions->debug,2,"%s\n",__func__);

  if (1<state) {
    pResult=StoreGet(pStore);
    pRequest=QueueLockWrite(pQueue,pResult,REQUEST_WRITE);
      pRequest->write.buf=buf;
      pRequest->write.len=len;
    QueueUnlockWrite(pQueue,pResult);
      failure=pResult->state;
    StorePut(pStore,pResult);
  }
  else
    failure=1;

  if (!failure)
    writtentime += len;

  return failure;
}

int canwrite()
{
  enum { SIZE=65536 };
  int available;
  Result *pResult;
  Request *pRequest;

  DPRINTF(pOptions->debug,2,"%s\n",__func__);

  if (1<state) {
    pResult=StoreGet(pStore);
    pRequest=QueueLockWrite(pQueue,pResult,REQUEST_CANWRITE);
    QueueUnlockWrite(pQueue,pResult);
      available=pResult->state;
    StorePut(pStore,pResult);
  }
  else
    available=0;

  return available;
}

int isplaying()
{
  int playing;
  Result *pResult;
  Request *pRequest;

  DPRINTF(pOptions->debug,0,"%s\n",__func__);

  if (1<state) {
    pResult=StoreGet(pStore);
    pRequest=QueueLockWrite(pQueue,pResult,REQUEST_ISPLAYING);
    QueueUnlockWrite(pQueue,pResult);
      playing=pResult->state;
    StorePut(pStore,pResult);
  }
  else
    playing=0;

  return playing;
}

int pause(int pause)
{
  int t=last_pause;
  Result *pResult;
  Request *pRequest;
  int time;

  DPRINTF(pOptions->debug,0,"%s\n",__func__);

  if (last_pause!=pause) {
    if (1<state) {
      pResult=StoreGet(pStore);
      pRequest=QueueLockWrite(pQueue,pResult,REQUEST_PAUSE);
        pRequest->pause.state=pause;
      QueueUnlockWrite(pQueue,pResult);
        time=pResult->state;
      StorePut(pStore,pResult);
    }
    else
      time=0;

    if (pause) {
      w_offset+=GETTIME(time)-start_t;
      writtentime=0;
    }
    else
      start_t=GETTIME(time);

    last_pause=pause;
  }

  return t;
}

void setvolume(int volume)
{
  if (0<state&&-666!=volume) {
    DPUTS(pOptions->debug,0,__func__ "\n");
    plugin.qVolume=(double)volume/255;
    DPRINTF(pOptions->debug,0,"  volume: %.2f\n",plugin.qVolume);
  }
}

void setpan(int pan)
{
}

void flush(int t)
{
  Result *pResult;
  Request *pRequest;
  int time;

  DPRINTF(pOptions->debug,0,"%s\n",__func__);

  if (1<state) {
    pResult=StoreGet(pStore);
    pRequest=QueueLockWrite(pQueue,pResult,REQUEST_FLUSH);
#if defined (YASAPI_STATE) // {
    pRequest->flush.time=t;
#endif // }
    QueueUnlockWrite(pQueue,pResult);
      time=pResult->state;
    StorePut(pStore,pResult);
  }
  else
    time=0;

  w_offset=t;
  start_t=GETTIME(time);
  writtentime=0;
}
  
int getoutputtime()
{
  if (state<2)
    return 0;
  else if (last_pause)
    return w_offset;
  else {
    DPRINTF(pOptions->debug,3,"%s\n",__func__);

    return GETTIME(GetAudioTime())-start_t + w_offset;
  }
}

int getwrittentime()
{
  int t=srate*numchan,l;
  int ms=writtentime;

  if (t) {
    l=ms%t;
    ms /= t;
    ms *= 1000;
    ms += (l*1000)/t;

    ms/=(bps/8);

    return ms + w_offset;
  }
  else
    return ms;
}

__declspec( dllexport ) Out_Module * winampGetOutModule()
{
  static Out_Module out;

  if (0==out.version) {
    out.version
      // module version (OUT_VER)
      =OUT_VER;
    out.description
      // description of module, with version string
      =PI_LABEL;
    out.id
      // module id. each input module gets its own.
      // non-nullsoft modules should be >= 65536. 
      =YASAPI_WA_ID;
    out.hMainWindow
      // winamp's main window (filled in by winamp)
      =NULL;
    out.hDllInstance
      // DLL instance handle (filled in by winamp)
      =NULL;
    out.Config
      // configuration dialog 
      =config;
    out.About
      // about dialog
      =about;
    out.Init
      // called when loaded
      =init;
    out.Quit
      // called when unloaded
      =quit;
    out.Open
      // returns >=0 on success, <0 on failure
      //
      // NOTENOTENOTE: bufferlenms and prebufferms are ignored in most
      // if not all output plug-ins. 
      //    ... so don't expect the max latency returned to be what you
      // asked for.
      // returns max latency in ms (0 for diskwriters, etc)
      // bufferlenms and prebufferms must be in ms. 0 to use defaults. 
      // prebufferms must be <= bufferlenms
      // pass bufferlenms==-666 to tell the output plugin that it's clock
      // is going to be used to sync video
      // out_ds turns off silence-eating when -666 is passed
      =open;
    out.Close
      // close the ol' output device.
      =close;
    out.Write
      // 0 on success. Len == bytes to write (<= 8192 always). buf is
      //   straight audio data. 
      // 1 returns not able to write (yet). Non-blocking, always.
      =write;
    out.CanWrite
      // returns number of bytes possible to write at a given time. 
      // Never will decrease unless you call Write (or Close, heh)
      =canwrite;
    out.IsPlaying
      // non0 if output is still going or if data in buffers waiting to be
      // written (i.e. closing while IsPlaying() returns 1 would truncate
      // the song
      =isplaying;
    out.Pause
      // returns previous pause state
      =pause;
    out.SetVolume
      // volume is 0-255
      =setvolume;
    out.SetPan
      // pan is -128 to 128
      =setpan;
    out.Flush
      // flushes buffers and restarts output at time t (in ms) 
      // (used for seeking)
      =flush;
    out.GetOutputTime
      // returns played time in MS
      =getoutputtime;
    out.GetWrittenTime
      // returns time written in MS (used for synching up vis stuff)
      =getwrittentime;
  }

  return &out;
}
