/*
 * yasapi.h
 * Copyright (C) 2015 Peter Belkner <pbelkner@snafu.de>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA
 */
#ifndef __YASAPI_H__ // {
#define __YASAPI_H__
#include <yasapi_guid.h>
#include <stdint.h>
#include <stdio.h>
#include <math.h>
#include <winamp/wa_ipc.h>
#include <winamp/out.h>
#ifdef __cpluplus
extern "C" {
#endif

///////////////////////////////////////////////////////////////////////////////
#define YASAPI_CONFIG2

///////////////////////////////////////////////////////////////////////////////
// win32 thread-safe queue implementation using native windows API
// http://stackoverflow.com/questions/11706985/win32-thread-safe-queue-implementation-using-native-windows-api

// Mutex vs. Semaphores – Part 1: Semaphores
// https://blog.feabhas.com/2009/09/mutex-vs-semaphores-%E2%80%93-part-1-semaphores/

// Mutex vs. Semaphores – Part 2: The Mutex
// https://blog.feabhas.com/2009/09/mutex-vs-semaphores-%E2%80%93-part-2-the-mutex/

// Mutex vs. Semaphores – Part 3 (final part): Mutual Exclusion Problems
// https://blog.feabhas.com/2009/10/mutex-vs-semaphores-%E2%80%93-part-3-final-part-mutual-exclusion-problems/

// Task Synchronisation
// https://blog.feabhas.com/2009/10/task-synchronisation/

// Task Synchronisation – Part 2: Multiple Tasks and RTOS APIs
// https://blog.feabhas.com/2009/11/task-synchronisation-part-2-multiple-tasks-and-rtos-apis/

///////////////////////////////////////////////////////////////////////////////
#if defined (_MSC_VER) && ! defined (__func__) // {
#define __func__                __FUNCTION__
#endif // }

// http://stackoverflow.com/questions/2192416/how-to-convert-concatenated-strings-to-wide-char-with-the-c-preprocessor
#define YASAPI_STR1(x)          #x
#define YASAPI_STR(x)           YASAPI_STR1(x)

#define YASAPI_WSTR1(x)         L ## #x
#define YASAPI_WSTR(x)          YASAPI_WSTR1(x)

// http://stackoverflow.com/questions/7795514/evaluate-preprocessor-token-before-concatenation
#define YASAPI_CONCAT1(A,B)     A ## B
#define YASAPI_CONCAT(A,B)      YASAPI_CONCAT1(A,B)

///////////////////////////////////////////////////////////////////////////////
#if ! defined (YASAPI_VER) // {
  #define YASAPI_VER            0.0.0
#endif // }

#define YASAPI_VERSION          YASAPI_STR(YASAPI_VER)
#define PI_VER2                 "v" YASAPI_VERSION

#if defined (OUT_YASAPI_SFX) // {
  #define YASAPI_PREFIX         YASAPI_CONCAT(out_yasapi,OUT_YASAPI_SFX)
#else // } {
  #define YASAPI_PREFIX         out_yasapi
#endif // }

#if defined (YASAPI_DEBUG) // {
  #define PI_VER                PI_VER2 " (Debug)"
  #define YASAPI_ID             YASAPI_PREFIX ## -debug
  #define YASAPI_WA_ID          (65536+855)
#elif defined (YASAPI_SSE2) // {
  #define PI_VER                PI_VER2 " (SSE2)"
  #define YASAPI_ID             YASAPI_PREFIX ## -sse2
  #define YASAPI_WA_ID          (65536+856)
#else // } {
  #define PI_VER                PI_VER2
  #define YASAPI_ID             YASAPI_PREFIX
  #define YASAPI_WA_ID          (65536+857)
#endif // }

#define YASAPI_NAME             YASAPI_STR(YASAPI_ID)
#if 0 // {
#define YASAPI_MODULE           YASAPI_NAME ".dll"
#endif // }
#define YASAPI_PROPERTY_FILE    L"Plugins\\" YASAPI_WSTR(YASAPI_ID) L".ini"
#define PI_LABEL                "Yet Another (WA)SAPI Output " PI_VER

#define YASAPI_INT_MAX(n)       (~(~0ll<<((n)-1)))

#define YASAPI_SHAREMODE_LABEL(eShareMode) \
    (AUDCLNT_SHAREMODE_SHARED==(eShareMode)?"SHARED":"EXCLUSIVE")

#if 0 // {
#if ! defined (YASAPI_AUDCLNT_SHAREMODE) // {
#define YASAPI_MODE_LABEL(mode) ( \
    YASAPI_MODE_SHARED==(mode)?"SHARED": \
    YASAPI_MODE_EXCLUSIVE==(mode)?"EXCLUSIVE": \
    "AUTOMATIC" \
)
#endif // }
#endif // }

///////////////////////////////////////////////////////////////////////////////
#if defined (YASAPI_DEBUG) // {
  #define GETDEBUG(debug)     (debug)
#else // } {
  #define GETDEBUG(debug)     0
#endif // }

#define DUNKNOWN(debug,y) \
  message(GETDEBUG(debug),1,0,&y,"UNKNOWN: \"%s\" (%d).\n", \
      basename(__FILE__),__LINE__)
#define DERROR(debug,x,y) \
  message(GETDEBUG(debug),0,x,&y,#x ": \"%s\" (%d).\n", \
      basename(__FILE__),__LINE__)
#define DERRORV(debug,x,y,format,...) \
  message(GETDEBUG(debug),0,x,&y,#x ": \"%s\" (%d).\n", \
      basename(__FILE__),__LINE__,__VA_ARGS__)

#if defined (YASAPI_DEBUG) // {
  #define DPUTS(debug,level,cs) \
      (((level)<(debug))&&fputs(cs,stderr))
  #define DPRINTF(debug,level,cs,...) \
      (((level)<(debug))&&fprintf(stderr,cs,__VA_ARGS__))
  #define DPUTWS(debug,level,ws) \
      (((level)<(debug))&&fputws(ws,stderr))
  #define DWPRINTF(debug,level,ws,...) \
      (((level)<(debug))&&fwprintf(stderr,ws,__VA_ARGS__))
  #define DMESSAGE(debug,m) (debug&&fprintf(stderr,"Error " m ": " \
      "\"%s\" (%d).\n",basename(__FILE__),__LINE__))
  #define DMESSAGEV(debug,m,...) (debug&&fprintf(stderr,"Error " m ": " \
      "\"%s\" (%d).\n",__VA_ARGS__,basename(__FILE__),__LINE__))
#else // } {
  #define DPUTS(debug,level,cs)
  #define DPRINTF(debug,level,cs,...)
  #define DPUTWS(debug,level,ws)
  #define DWPRINTF(debug,level,ws,...)
  #define DMESSAGE(debug,m)
  #define DMESSAGEV(debug,m,...)
#endif // }

///////////////////////////////////////////////////////////////////////////////
#define YASAPI_FRAMES_MS(nFrames,pwfx) \
  MulDiv(1000,nFrames,(pwfx)->nSamplesPerSec)
#define YASAPI_FRAMES_TIMER(nFrames,pwfx) \
  (10000.0*1000*(nFrames)/(pwfx)->nSamplesPerSec)

#define YASAPI_MIN_BALANCE_COUNT  (8ull)

#define YASAPI_MIN_RING_SIZE      (1.0)
#define YASAPI_MAX_RING_SIZE      (7.0)

#define YASAPI_MIN_SHARE_SIZE     (1.0)
#define YASAPI_MAX_SHARE_SIZE     (20.0)

///////////////////////////////////////////////////////////////////////////////
DEFINE_GUID(IID_IUnknown, 0x00000000, 0x0000, 0x0000, 0xc0,0x00,
    0x00,0x00,0x00,0x00,0x00,0x46);

DEFINE_GUID(IID_IAudioClient, 0x1cb9ad4c, 0xdbfa, 0x4c32, 0xb1,0x78,
    0xc2,0xf5,0x68,0xa7,0x03,0xb2);
DEFINE_GUID(IID_IAudioClock, 0xcd63314f, 0x3fba, 0x4a1b, 0x81,0x2c,
    0xef,0x96,0x35,0x87,0x28,0xe7);
DEFINE_GUID(IID_IAudioRenderClient, 0xf294acfc, 0x3146, 0x4483, 0xa7,0xbf,
    0xad,0xdc,0xa7,0xc2,0x60,0xe2);

DEFINE_GUID(CLSID_MMDeviceEnumerator, 0xbcde0395, 0xe52f, 0x467c, 0x8e,0x3d,
    0xc4,0x57,0x92,0x91,0x69,0x2e);
DEFINE_GUID(IID_IMMDeviceEnumerator, 0xa95664d2, 0x9614, 0x4f35, 0xa7,0x46,
    0xde,0x8d,0xb6,0x36,0x17,0xe6);
DEFINE_GUID(IID_IMMDevice, 0xd666063f, 0x1587, 0x4e43, 0x81,0xf1,
    0xb9,0x48,0xe8,0x07,0x36,0x3f);

///////////////////////////////////////////////////////////////////////////////
typedef struct _PropertyType PropertyType;
typedef struct _PropertyLoadConfig PropertyLoadConfig;
typedef struct _PropertySaveConfig PropertySaveConfig;
typedef struct _Property Property;
typedef struct _SessionCreateConfig SessionCreateConfig;
typedef struct _SessionWriteConfig SessionWriteConfig;
typedef struct _SessionPlayConfig SessionPlayConfig;
typedef struct _Result Result;
typedef struct _Store Store;
typedef struct _Request Request;
typedef struct _QueueStrategy QueueStrategy;
typedef struct _Queue Queue;
typedef struct _Ring Ring;
typedef struct _Strategy Strategy;
typedef struct _Session Session;
typedef struct _ChannelOptions ChannelOptions;
typedef struct _Channel Channel;
typedef struct _Player Player;
typedef struct _PluginOptions PluginOptions;
typedef struct _Plugin Plugin;

typedef struct _ControlCheckBoxConfig ControlCheckBoxConfig;
typedef struct _ControlRadioButtonConfig ControlRadioButtonConfig;
typedef struct _ControlSliderConfig ControlSliderConfig;
typedef struct _ControlComboBoxList ControlComboBoxList;
typedef struct _ControlComboBoxConfig ControlComboBoxConfig;
typedef struct _ControlSliderCascadeList ControlSliderCascadeList;
typedef struct _ControlSliderCascadeConfig ControlSliderCascadeConfig;

typedef struct _ControlType ControlType;
typedef struct _Control Control;

///////////////////////////////////////////////////////////////////////////////
__declspec( dllexport ) Out_Module * winampGetOutModule();
char *basename(char *s);
wchar_t *wapath(wchar_t *file);
int message(int debug, int force, HRESULT x, HRESULT *y, const char *m, ...);
int crash(const char *format, ...);
#if defined (YASAPI_DEBUG) // {
void ConsoleOpen(void);
void ConsoleClose(void);
#endif // }

///////////////////////////////////////////////////////////////////////////////
#define VALUE_INT(pData,offs) \
  (*(int *)(((char *)(pData))+(offs)))
#define VALUE_DOUBLE(pData,offs) \
  (*(double *)(((char *)(pData))+(offs)))
#define VALUE_STRING(pData,offs) \
  ((wchar_t *)(((char *)(pData))+(offs)))

#define PROPERTY_INT(pProperty,pData) \
  VALUE_INT(pData,(pProperty)->offs)
#define PROPERTY_DOUBLE(pProperty,pData) \
  VALUE_DOUBLE(pData,(pProperty)->offs)
#define PROPERTY_STRING(pProperty,pData) \
  VALUE_STRING(pData,(pProperty)->offs)

#define YASAPI_PROPERTY_SIZE    128
#define YASAPI_GROUP_GENERAL    L"plugin"
#define YASAPI_PROPERTY_DEBUG   L"debug"
#define YASAPI_PROPERTY_VIS     L"visualization"
#define YASAPI_PROPERTY_PAGE    L"page"
#define YASAPI_PROPERTY_POSX    L"posx"
#define YASAPI_PROPERTY_POSY    L"posy"

struct _PropertyLoadConfig {
  int debug;
  const wchar_t *path;
  void *pData;
  const void *pDefault;
  const wchar_t *group;
  const wchar_t *pfx;
};

struct _PropertySaveConfig {
  const wchar_t *path;
  void *pData;
  const wchar_t *group;
};

struct _PropertyType {
  void (*Load)(const Property *pProperty, PropertyLoadConfig *c);
  void (*Save)(const Property *pProperty, const PropertySaveConfig *c);
};

struct _Property {
  wchar_t *key;
  const PropertyType *type;
  int offs;
  int size;
};

extern const PropertyType gcIntType;
extern const PropertyType gcDoubleType;
extern const PropertyType gcStringType;

void PropertySaveInt(const wchar_t *group, const wchar_t *key, int n,
    const wchar_t *path);

void PropertyLoad(const Property *pProperty, PropertyLoadConfig *c);
void PropertySave(const Property *pProperty, const PropertySaveConfig *c);

///////////////////////////////////////////////////////////////////////////////
struct _Result {
  Result *prPrev;
  HANDLE hEvent;

  union {
    int state;
  };
};

int ResultCreate(Result *pResult, int debug);
void ResultDestroy(Result *pResult, int debug);

///////////////////////////////////////////////////////////////////////////////
enum {
  STORE_SIZE=12
};

struct _Store {
  Result *prTos;
  Result aResult[STORE_SIZE];
  HANDLE hAvailable;
  HANDLE hMutex;
#if defined (YASAPI_DEBUG) // {
  int nAvailable;
  int nMinAvailable;
#endif // }
};

int StoreCreate(Store *pStore, int debug);
void StoreDestroy(Store *pStore, int debug);

void StorePush(Store *pStore, Result *pResult);
Result *StorePop(Store *pStore);

void StoreLock(Store *pStore);
void StoreUnlock(Store *pStore);

Result *StoreGet(Store *pStore);
void StorePut(Store *pStore, Result *pResult);

///////////////////////////////////////////////////////////////////////////////
enum {
  REQUEST_STOP,
  REQUEST_CREATE,
  REQUEST_DESTROY,
  REQUEST_OPEN,
  REQUEST_CLOSE,
  REQUEST_WRITE,
  REQUEST_READ,
  REQUEST_PAUSE,
  REQUEST_FLUSH,
  REQUEST_CANWRITE,
  REQUEST_ISPLAYING,
  REQUEST_GETTIME
};

struct _SessionCreateConfig {
  int samplerate;
  int numchannels;
  int bitspersamp;
};

struct _SessionWriteConfig {
  const char *buf;
  int len;
};

struct _Request {
  Result *pResult;
  int tag;

  union {
    SessionCreateConfig open;
    SessionWriteConfig write;

    struct {
      int state;
    } pause;

    struct {
      int fullstop;
    } close;

    struct {
      int time;
    } flush;
  };
};

///////////////////////////////////////////////////////////////////////////////
struct _QueueStrategy {
  const char *name;
  HANDLE (*GetSemaphoreDown)(Queue *pQueue);
  HANDLE (*GetSemaphoreUp)(Queue *pQueue);
  BOOL (*IsAlertable)(Queue *pQueue);
  Request **(*GetRequest)(Queue *pQueue);
#if defined (YASAPI_DEBUG) // {
  void (*Inc)(Queue *pQueue);
  void (*Dec)(Queue *pQueue);
#endif // }
};

extern const QueueStrategy cqsWrite;
extern const QueueStrategy cqsRead;

///////////////////////////////////////////////////////////////////////////////
enum {
  QUEUE_SIZE=12
};

struct _Queue {
  Request aRequest[QUEUE_SIZE];
  Request *rp;
  Request *wp;
  Request *mp;
  HANDLE hAvailable;
  HANDLE hWritten;
  HANDLE hMutex;
#if defined (YASAPI_DEBUG) // {
  int nAvailable;
  int nMinAvailable;
#endif // }
};

int QueueCreate(Queue *pQueue, int debug);
void QueueDestroy(Queue *pQueue, int debug);

void QueueLockMutex(Queue *pQueue);
void QueueUnlockMutex(Queue *pQueue);

Request *QueueLock(Queue *pQueue, const QueueStrategy *pStrategy,
    HANDLE hEvent, void (*OnEvent)(void *data), void *data);
void QueueUnlock(Queue *pQueue, const QueueStrategy *pStrategy,
    Result *pResult);
Request *QueueLockWrite(Queue *pQueue, Result *pResult, int tag);
Request *QueueLockRead(Queue *pQueue, HANDLE hEvent,
    void (*OnEvent)(void *data), void *data);
void QueueUnlockRead(Queue *pQueue, Result *pResult);
void QueueWrite(Queue *pQueue, int tag, Store *pStore);

#if defined (YASAPI_DEBUG) // {
void QueueIncAvailable(Queue *pQueue);
void QueueDecAvailable(Queue *pQueue);
void QueueNopAvailable(Queue *pQueue);
#endif // }

///////////////////////////////////////////////////////////////////////////////
typedef void (*RingCopy)(PVOID pClient, PVOID Destination,
    const VOID *Source, SIZE_T Length);

enum {
  RING_COPY=1,
  RING_COMMIT=2
};

struct _Ring {
  RingCopy Copy;
  PVOID pClient;
  LPSTR pRep;
  LPSTR pMax;
  LPSTR pHead;
  LPSTR pTail;
  SIZE_T dwWritten;
  SIZE_T dwAvailable;
};

DWORD RingCreate(Ring *pRing, SIZE_T dwSize, int debug);
void RingDestroy(Ring *pRing);

void RingReset(Ring *pRing);
SIZE_T RingWrite(Ring *pRing, LPCSTR pData, SIZE_T dwSize, SIZE_T n);
SIZE_T RingReadEx(Ring *pRing, LPSTR pData, SIZE_T dwSize, UINT uFlags);
SIZE_T RingRead(Ring *pRing, LPSTR pData, SIZE_T dwSize);

void RingCopyMemory(PVOID pClient, PVOID Destination, const VOID *Source,
    SIZE_T Length);

///////////////////////////////////////////////////////////////////////////////
struct _Strategy {
  HRESULT (*Initialize)(Session *pSession, AUDCLNT_SHAREMODE eShareMode,
      DWORD dwStreamFlags, REFERENCE_TIME hnsDevicePeriod);
  void (*BalanceStart)(Session *pSession);
  void (*BalanceIterate)(Session *pSession, UINT32 uFramesWrite);
  void (*SetTimer)(Session *pSession, LONGLONG time, BOOL bFlush);
  BOOL (*NeedDevicePeriod)(Session *pSession, AUDCLNT_SHAREMODE eShareMode);
  BOOL (*CanAdapt)(Session *pSession);
  BOOL (*IsPush)(Session *pSession);
  UINT32 (*GetFramesPadding)(Session *pSession);
};

extern const Strategy gcsPush;
extern const Strategy gcsPull;

///////////////////////////////////////////////////////////////////////////////
#define SESSION_IS_RUNNING(pSession) \
  ((pSession)->state.started&&!(pSession)->state.pause)

struct _SessionPlayConfig {
  UINT32 uFramesPadding;
  UINT32 uFramesWrite;
  UINT32 uFramesRing;
  UINT32 uSizePad;
  BOOL bFlush;
};

struct _Session {
  Channel *pChannel;
  const Strategy *pStrategy;
  AUDCLNT_SHAREMODE eShareMode;
  WAVEFORMATEXTENSIBLE wfxx;
  WORD wBytesPerSample;
  BOOL bMono2Stereo;

  UINT32 uFrames;
  UINT32 uFramesMin;
  UINT32 uFramesPadding;
  IAudioClient *pClient;
  IAudioClock *pClock;
  IAudioRenderClient *pRender;
  Ring ring;
  LONGLONG time;

  struct {
    double q;
    ULONG64 count;
  } balance;

  struct {
    int valid:1;
    int started:1;
    int pause:1;
    int wait:1;
    int drain:1;
    int flush:1;
    int eof:1;
  } state;
};

int SessionCreate(Session *pSession, const SessionCreateConfig *pConfig,
    Channel *pChannel);
void SessionDestroy(Session *pSession);

Plugin *SessionGetPlugin(Session *pSession);
int SessionGetDebug(Session *pSession);
int SessionCanWrite(Session *pSession);

void SessionPostUpdate(Session *pSession, SessionPlayConfig *c);

int SessionStart(Session *pSession);
void SessionStop(Session *pSession, BOOL bFlush);

UINT32 SessionGetFramesPadding(Session *pSession);
int SessionWrite(Session *pSession, const SessionWriteConfig *pConfig);
int SessionWriteNext(Session *pSession, SessionWriteConfig *c);
void SessionPlayConfigSetup(SessionPlayConfig *c, UINT32 uFramesPadding);
int SessionWriteBuffer(Session *pSession, SessionPlayConfig *c);
int SessionPlay(Session *pSession, SessionPlayConfig *c);
int SessionRead(Session *pSession);
int SessionGetTime(Session *pSession);

void SessionSendRead(void *data);
VOID CALLBACK SessionFireTimer(LPVOID lpArg, DWORD dwLow, DWORD dwHigh);
void SessionSetTimer(Session *pSession, LONGLONG time);

void SessionCopyMemory(PVOID pClient, PVOID Destination, const VOID *Source,
    SIZE_T Length);

///////////////////////////////////////////////////////////////////////////////
enum {
  YASAPI_SHAREMODE_SHARE,
  YASAPI_SHAREMODE_EXCLUSIVE,
  YASAPI_SHAREMODE_AUTOMATIC
};

enum {
  YASAPI_DEVICE_PERIOD_DEFAULT,
  YASAPI_DEVICE_PERIOD_MINIMUM,
};

struct _ChannelOptions {
  Channel *pChannel;
  // Persistent fields have to be of type "int", "double" or "wchar_t[]".
  int eShareMode;
  int eDevicePeriod;
  int bAutoConvertPCM;
  int bSRCDefaultQuality;
  int bPull;
  int bBalance;
  double qBalanceStart;
  double qShareSize;
  double qRingSize;
  double qRingFill;
};

int ChannelOptionsCreate(ChannelOptions *pOptions, Channel *pChannel);
void ChannelOptionsDestroy(ChannelOptions *pOptions);

void ChannelOptionsSave(ChannelOptions *pOptions);
const ChannelOptions *ChannelOptionsDefault(void);

///////////////////////////////////////////////////////////////////////////////
struct _Channel {
  Player *pPlayer;
  IMMDevice *pDevice;
  LPWSTR pstrId;
  IPropertyStore *pProperties;
  PROPVARIANT vName;
  ChannelOptions options;
};

int ChannelCreate(Channel *pChannel, Player *pPlayer, UINT uDevice);
void ChannelDestroy(Channel *pChannel);

///////////////////////////////////////////////////////////////////////////////
struct _Player {
  Plugin *pPlugin;
  IMMDeviceEnumerator *pEnumerator;
  IMMDeviceCollection *pCollection;
  UINT cDevices;
  Channel *pChannel;
  UINT uDevice;
};

int PlayerCreate(Player *pPlayer, Plugin *pPlugin);
void PlayerDestroy(Player *pPlayer);

HWND PlayerGetDlgConfig(Player *pPlayer);
void PlayerPostUpdate(Player *pPlayer, WORD wRing, WORD wShared);

///////////////////////////////////////////////////////////////////////////////
enum {
  YASAPI_ID_SIZE=128
};

struct _PluginOptions {
  const wchar_t *path;
  // Persistent fields have to be of type "int", "double" or "wchar_t[]".
  int debug;
  wchar_t id[YASAPI_ID_SIZE];
  int bMono2Stereo;
  int bVolume;
  int bWriteBlock;
  int bAudioClock;
  int bVisualization;
  int page;
  int posx;
  int posy;
#if defined (YASAPI_DEBUG) // {
  int sleep;
#endif // }
};

int PluginOptionsCreate(PluginOptions *pOptions, const wchar_t *path,
    int debug);
void PluginOptionsDestroy(PluginOptions *pOptions);

void PluginOptionsSave(PluginOptions *pOptions);
const PluginOptions *PluginOptionsDefault(void);

///////////////////////////////////////////////////////////////////////////////
struct _Plugin {
  PluginOptions options;
  Queue queue;
  Store store;
  double qVolume;
  Player *pPlayer;
  HWND hDlgConfig;
  HANDLE hTimer;
  HANDLE hEvent;
  HANDLE hThread;
};

int PluginCreate(Plugin *pPlugin, const wchar_t *path, int debug);
void PluginDestroy(Plugin *pPlugin);

DWORD WINAPI PluginThread(LPVOID lpParameter);

///////////////////////////////////////////////////////////////////////////////
struct _ControlType {
  void (*Init)(const Control *pControl, HWND hDlg, HWND hWnd);
  void (*Set)(const Control *pControl, HWND hDlg, const void *pData);
  void (*Get)(const Control *pControl, HWND hDlg, void *pData);
  void (*Sync)(const Control *pControl, HWND hDlg);
};

struct _ControlCheckBoxConfig {
  int idc;
  int offset;
  int off;
  int on;
  wchar_t *help;
};

struct _ControlRadioButtonConfig {
  int idc;
  int offset;
  int val;
  wchar_t *help;
};

struct _ControlSliderConfig {
  int idcSlider;
  int offset;
  double min;
  double max;
  LONG lMaxRange;
  int idcStatic;
  const wchar_t *format;
  wchar_t *help;
};

struct _ControlComboBoxList {
  const wchar_t *label;
  int val;
};

struct _ControlComboBoxConfig {
  int idc;
  int offset;
  const ControlComboBoxList *list;
  wchar_t *help;
};

struct _ControlSliderCascadeList {
  int idcSlider;
  int offset;
  int idcStatic;
  const wchar_t *format;
  wchar_t *help;
};

struct _ControlSliderCascadeConfig {
  double min;
  double max;
  LONG lMaxRange;
  const ControlSliderCascadeList *list;
};

struct _Control {
  const ControlType *type;
  const void *config;
};

extern const ControlType gcCheckBoxType;
extern const ControlType gcRadioButtonType;
extern const ControlType gcSliderType;
extern const ControlType gcComboBoxType;
extern const ControlType gcSliderCascadeType;

HWND ControlCreateToolTip(HWND hDlg, HINSTANCE hInstance);
void ControlAddToolTip(HWND hDlg, HWND hWnd, int idc, PWSTR pszText);

///////////////////////////////////////////////////////////////////////////////
extern const Control gcaPluginControlList[];
extern const Control gcaDeviceControlList[];
#if defined (YASAPI_CONFIG2) // {
extern const Control gcaSimplifiedDeviceControlList[];
#endif // }
extern const Control gcaBuffersControlList[];

void ControlInit(const Control *pControl, HWND hDlg, HINSTANCE hInstance);
void ControlSet(const Control *pControl, HWND hDlg, const void *pData);
void ControlGet(const Control *pControl, HWND hDlg, void *pData);
void ControlSync(const Control *pControl, HWND hDlg);

///////////////////////////////////////////////////////////////////////////////
enum {
  WM_CONFIG_UPDATE=WM_APP
};

void ConfigDialog(HWND hWndParent, HINSTANCE hInstance, Player *pPlayer);
void AboutDialog(HWND hWndParent);

#ifdef __cpluplus
}
#endif
#endif // }
