/*
 * yasapi_channel.c
 * Copyright (C) 2015 Peter Belkner <pbelkner@snafu.de>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA
 */
#include <yasapi.h>
#if ! defined (PKEY_Device_FriendlyName) // {
#include <FunctionDiscoveryKeys_devpkey.h>
#endif // }

int ChannelCreate(Channel *pChannel, Player *pPlayer, UINT uDevice)
{
#if defined (YASAPI_DEBUG) // {
  int debug=pPlayer->pPlugin->options.debug;
#endif // }
  IMMDeviceCollection *pCollection=pPlayer->pCollection;
  IMMDevice *pDevice;
  LPWSTR pstrId;
  IPropertyStore *pProperties;
  HRESULT hr;

  /////////////////////////////////////////////////////////////////////////////
  pChannel->pPlayer=pPlayer;

  // get the player device ////////////////////////////////////////////////////
  hr=pCollection->lpVtbl->Item(pCollection,
    uDevice,                  // UINT      nDevice,
    &pDevice                  // IMMDevice **ppDevice
  );

  if (FAILED(hr)) {
    DERROR(debug,E_POINTER,hr);
    DERROR(debug,E_INVALIDARG,hr);
    DERROR(debug,E_OUTOFMEMORY,hr);
    DMESSAGE(debug,"getting player device");
    goto device;
  }

  pChannel->pDevice=pDevice;
  DPUTS(debug,0,"    player device created\n");

  // get the device's id //////////////////////////////////////////////////////
  hr=pDevice->lpVtbl->GetId(pDevice,
    &pstrId                   // [out] LPWSTR *ppstrId
  );

  if (FAILED(hr)) {
    DERROR(debug,E_OUTOFMEMORY,hr);
    DERROR(debug,E_POINTER,hr);
    DMESSAGE(debug,"getting player device id");
    goto id;
  }

  pChannel->pstrId=pstrId;
  DWPRINTF(debug,0,L"    id: \"%s\"\n",pChannel->pstrId);
  
  // getting the device's property store //////////////////////////////////////
  hr=pDevice->lpVtbl->OpenPropertyStore(pDevice,
    STGM_READ,              // [in]  DWORD          stgmAccess,
    &pProperties            // [out] IPropertyStore **ppProperties
  );

  if (FAILED(hr)) {
    DERROR(debug,E_INVALIDARG,hr);
    DERROR(debug,E_POINTER,hr);
    DERROR(debug,E_OUTOFMEMORY,hr);
    DMESSAGE(debug,"getting player property store");
    goto properties;
  }

  DPUTS(debug,0,"    player property store created\n");
  pChannel->pProperties=pProperties;

  // getting the friendly name of the audio adapter to which the //////////////
  // endpoint device is attached
  PropVariantInit(&pChannel->vName);

  hr=pProperties->lpVtbl->GetValue(pProperties,
    &PKEY_Device_FriendlyName,
                            // [in]  REFPROPERTYKEY key,
    &pChannel->vName        // [out] PROPVARIANT    *pv
  );

  if (FAILED(hr)) {
    DMESSAGE(debug,"retrieving PKEY_Device_FriendlyName property");
    goto name;
  }

  DWPRINTF(debug,0,L"    name: \"%s\"\n",pChannel->vName.pwszVal);

  // options //////////////////////////////////////////////////////////////////
  if (ChannelOptionsCreate(&pChannel->options,pChannel)<0) {
    DMESSAGE(debug,"creating player options");
    goto options;
  }

  DPUTS(debug,0,"    player options created\n");

  // success //////////////////////////////////////////////////////////////////
  return 0;
//cleanup:
  //ChannelOptionsDestroy(&pChannel->options);
options:
  PropVariantClear(&pChannel->vName);
name:
  pProperties->lpVtbl->Release(pProperties);
properties:
  CoTaskMemFree(pstrId);
id:
  pDevice->lpVtbl->Release(pDevice);
device:
  return -1;
}

void ChannelDestroy(Channel *pChannel)
{
#if defined (YASAPI_DEBUG) // {
  int debug=pChannel->pPlayer->pPlugin->options.debug;
#endif // }
  IMMDevice *pDevice=pChannel->pDevice;
  IPropertyStore *pProperties=pChannel->pProperties;

  /////////////////////////////////////////////////////////////////////////////
  DPUTS(debug,0,"    destroying player options\n");
  ChannelOptionsDestroy(&pChannel->options);

  DPUTS(debug,0,"    deleting player device name\n");
  PropVariantClear(&pChannel->vName);

  DPUTS(debug,0,"    destroying player property store\n");
  pProperties->lpVtbl->Release(pProperties);

  DPUTS(debug,0,"    deleting player device id\n");
  CoTaskMemFree(pChannel->pstrId);

  DPUTS(debug,0,"    destroying player device\n");
  pDevice->lpVtbl->Release(pDevice);

  /////////////////////////////////////////////////////////////////////////////
  pChannel->pPlayer=NULL;
}
